/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.awt.Shape;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import jp.digitalmuseum.mr.core.message.MrEvent;
import jp.digitalmuseum.mr.core.message.MrEventListener;

/**
 * Abstract implementation of ProxyRobot.<br />
 * VirtualRobot implementation classes must extend this abstract class.
 *
 * @author Jun KATO
 */
public abstract class ProxyRobotAbstractImpl implements ProxyRobot {

	/** Name of this robot. */
	private String name;

	/** Name of the type of this robot. */
	private String typeName;

	/** Set of resources in use */
	private HashMap<Resource, Behavior> inUse;

	/** List of MrEventListener */
	private List<MrEventListener> listeners;

	public ProxyRobotAbstractImpl() {
		inUse = new HashMap<Resource, Behavior>();
		listeners = new ArrayList<MrEventListener>();
		typeName = getClass().getSimpleName();
		ProxyRobotHolder.getInstance().registerEntity(this);
	}

	public void dispose() {
		ProxyRobotHolder.getInstance().unregisterEntity(this);
	}

	public Shape getShape() {
		return null;
	}

	/**
	 * Set a name of this robot.<br />
	 *
	 * @param name
	 */
	final protected void setName(String name) {
		this.name = name;
	}

	final public String getName() {
		return name;
	}

	/**
	 * Set a name of the type of this robot.<br />
	 * <b>Caution:</b> This method should only be called by a subclass of VirtualRobotAbstractImpl.
	 *
	 * @param typeName
	 */
	final protected void setTypeName(String typeName) {
		this.typeName = typeName;
	}

	final public String getTypeName() {
		return typeName;
	}

	/**
	 * This method must be implemented by child classes.<br />
	 * <b>Caution:</b> Do not call this method outside this class.
	 *
	 * @return Returns an array of resources.
	 */
	protected abstract Resource[] getResources();

	final public Set<Class<? extends Resource>> getResourceTypes() {
		final HashSet<Class<? extends Resource>> set =
				new HashSet<Class<? extends Resource>>();

		// Find out types of every resource
		for (Resource resource : getResources()) {
			set.add(resource.getClass());
		}
		return set;
	}

	final synchronized public Set<Class<? extends Resource>> getAvailableResourceTypes() {
		final HashSet<Class<? extends Resource>> set =
			new HashSet<Class<? extends Resource>>();

		// Find out types of every resource
		for (Resource resource : getResources()) {
			if (!inUse.keySet().contains(resource)) {
				set.add(resource.getClass());
			}
		}
		return set;
	}

	final synchronized public Map<Class<? extends Resource>, Behavior> getBehaviorMap() {
		HashMap<Class<? extends Resource>, Behavior> behaviorMap =
			new HashMap<Class<? extends Resource>, Behavior>();
		for (Entry<Resource, Behavior> entry : inUse.entrySet()) {
			behaviorMap.put(entry.getKey().getClass(), entry.getValue());
		}
		return behaviorMap;
	}

	final synchronized public Behavior getAssignedBehavior(Class<? extends Resource> resource) {
		for (Entry<Resource, Behavior> entry : inUse.entrySet()) {
			if (CoreUtils.objectCastableTo(entry.getKey(), resource)) {
				return entry.getValue();
			}
		}
		return null;
	}

	/**
	 * Child class should override this method to request resources of related robots.
	 */
	public synchronized HashMap<Class<? extends Resource>, Resource>
			requestResources(Behavior behavior, Set<Class<? extends Resource>> resourceTypes) {

		// Get a set of free resources.
		final HashSet<Resource> freeResources =
			new HashSet<Resource>(Arrays.asList(getResources()));
		freeResources.removeAll(inUse.keySet());

		// Construct a resource map.
		final HashMap<Class<? extends Resource>, Resource> resourceMap =
			CoreUtils.constructResourceMap(resourceTypes, freeResources);

		// Set "in use" flags if succeeded.
		if (resourceMap != null) {
			for (Resource resource : resourceMap.values()) {
				inUse.put(resource, behavior);
			}
		}

		// Returns a resource map.
		return resourceMap;
	}

	/**
	 * Child class should override this method to free resources of related robots.
	 */
	protected synchronized void freeResource(Resource resource) {

		// Unset "in use" flag
		inUse.remove(resource);
	}

	/**
	 * Bring the given event to the event listeners.<br />
	 * <b>Caution:</b> This method should only be called by a subclass of RobotAbstractImpl.
	 *
	 * @param event event to be brought to listeners.
	 * @see #addEventListener(MrEventListener)
	 * @see #removeEventListener(MrEventListener)
	 */
	protected void distributeRobotEvent(MrEvent event) {
		synchronized (listeners) {
			for (MrEventListener listener : listeners) {
				listener.eventOccurred(event);
			}
		}
	}

	public void addEventListener(MrEventListener mrEventListener) {
		synchronized (listeners) {
			listeners.add(mrEventListener);
		}
	}

	public boolean removeEventListener(MrEventListener mrEventListener) {
		synchronized (listeners) {
			return listeners.remove(mrEventListener);
		}
	}

}
