/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import jp.digitalmuseum.mr.core.message.MrEvent;
import jp.digitalmuseum.mr.core.message.MrEventListener;
import jp.digitalmuseum.mr.core.message.MrServiceEvent;

/**
 * Abstract implementation for Service interface.<br />
 * All Service implementation classes must extend this abstract class.
 *
 * @author Jun KATO
 */
public abstract class ServiceAbstractImpl implements Service {
	private static long id = 0;
	private long myId = id ++;
	private ServiceGroup serviceGroup;
	private Map<Object, String> catalogEntry;
	private ServiceSettingHandler serviceSettingHandler;
	final private ArrayList<MrEventListener> listeners;
	private int listenersCounter;
	private long birthDate;
	private boolean isPaused;

	public ServiceAbstractImpl() {
		catalogEntry = new HashMap<Object, String>();
		listeners = new ArrayList<MrEventListener>();
	}

	synchronized public void start() {
		registerTo(Matereal.getInstance());
	}

	synchronized public void start(ServiceGroup serviceGroup) {
		registerTo(serviceGroup);
	}

	/**
	 * Register this service to the specified group.
	 *
	 * @param group
	 */
	void registerTo(ServiceGroup serviceGroup) {
		if (this.serviceGroup != null) {
			stop();
		}

		// Register this service.
		this.serviceGroup = serviceGroup;
		serviceGroup.registerService(this);

		// Save the date of birth.
		birthDate = System.currentTimeMillis();

		// Distribute this event.
		distributeEvent(
				new MrServiceEvent(
						ServiceAbstractImpl.this,
						MrServiceEvent.STATUS.STARTED));
	}

	public void pause() {
		isPaused = true;

		// Distribute this event.
		distributeEvent(
				new MrServiceEvent(
						ServiceAbstractImpl.this,
						MrServiceEvent.STATUS.PAUSED));
	}

	public void resume() {
		isPaused = false;

		// Distribute this event.
		distributeEvent(
				new MrServiceEvent(
						ServiceAbstractImpl.this,
						MrServiceEvent.STATUS.RESUMED));
	}

	synchronized public void stop() {

		if (!isStarted()) {
			// throw new IllegalStateException("The service is not yet started.");
			return;
		}

		// Unregister this service.
		if (!serviceGroup.isDisposing()) {
			serviceGroup.unregisterService(this);
		}
		serviceGroup = null;

		// Reset the born time of this service.
		birthDate = 0;

		// Distribute this event.
		distributeEvent(
				new MrServiceEvent(
						ServiceAbstractImpl.this,
						MrServiceEvent.STATUS.STOPPED));
	}

	public void dispose() {
		// Do nothing by default.
	}

	public boolean isStarted() {
		return serviceGroup != null;
	}

	public boolean isPaused() {
		return isPaused;
	}

	public long getId() {
		return myId;
	}

	public long getInterval() {
		return isStarted() ? serviceGroup.getInterval() : 0;
	}

	public ServiceGroup getServiceGroup() {
		return serviceGroup;
	}

	public Map<Object, String> getCatalogEntry() {
		return catalogEntry;
	}

	public ServiceSettingHandler getDialogHandler() {
		// TODO implement catalog related codes.
		if (serviceSettingHandler == null) {
			serviceSettingHandler = new ServiceSettingHandlerImpl();
		}
		return serviceSettingHandler;
	}

	public String getName() {
		return "";
	}

	public synchronized long getAliveTime() {
		return birthDate > 0 ? System.currentTimeMillis() - birthDate : -1;
	}

	@Override
	public String toString() {
		return getName();
	}

// Event related methods.
// TODO think of using VLA instead of ArrayList

	/**
	 * Add an event listener.
	 */
	public void addEventListener(MrEventListener listener) {
		synchronized (listeners) {
			listeners.add(listener);
		}
	}

	/**
	 * Remove an event listener.
	 *
	 * @return Returns whether removal succeeded or not.
	 */
	public boolean removeEventListener(MrEventListener listener) {
		if (listener == null) {
			return false;
		}
		synchronized (listeners) {
			for (int i = 0; i < listeners.size(); i ++) {
				if (listener.equals(listeners.get(i))) {
					if (i <= listenersCounter) {
						listenersCounter --;
					}
					listeners.remove(i);
					return true;
				}
			}
			return false;
		}
	}

	/**
	 * Distribute an event to listeners.
	 *
	 * @param e
	 */
	protected void distributeEvent(MrEvent e) {
		synchronized (listeners) {
			for (listenersCounter = 0; listenersCounter < listeners.size(); listenersCounter ++) {
				final MrEventListener listener = listeners.get(listenersCounter);
				listener.eventOccurred(e);
			}
		}
	}
}
