/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.awt.event.ActionListener;
import java.util.List;


/**
 * Interface for managing periodic tasks.
 *
 * @author Jun KATO
 * @see Matereal
 * @see ServiceGroupImpl
 */
public interface ServiceGroup extends Iterable<Service>, Disposable {

	/**
	 *
	 */
	public abstract String getName();

	/**
	 * Set duration of one cycle in milliseconds.
	 * @param period Duration of a period in milliseconds.
	 */
	public abstract void setInterval(long period);

	/**
	 * Get duration of one cycle in milliseconds.
	 * @return Returns the duration
	 */
	public abstract long getInterval();

	/**
	 * Add an ActionListener to draw something,
	 * given the same interval with this service group.<br />
	 * Specified ActionListener is registered to the Swing Timer object inside this service group.
	 *
	 * @see #removeActionListener(ActionListener)
	 */
	public abstract void addActionListener(ActionListener listener);

	/**
	 * Remove an ActionListner from the Swing Timer object inside this service group.
	 *
	 * @see #addActionListener(ActionListener)
	 */
	public abstract void removeActionListener(ActionListener listener);

	/**
	 * Register given service to run in this service group.
	 * @param service
	 */
	public abstract void registerService(Service service);

	/**
	 * Unregister given service to stop running in this service group.
	 * @param service Unregistered service when succeeded in unregistration.
	 */
	public abstract Service unregisterService(Service service);

	/**
	 * Unregister all services related to this service group.
	 * @return Returns all unregistered services.
	 */
	public abstract List<Service> clearServices();

	/**
	 * Get all services related to this service group.
	 * @return
	 */
	public abstract List<Service> getServices();

	/**
	 * Merge other service group with this group.<br />
	 * Merged group will be disposed using dispose().
	 */
	public abstract void mergeServiceGroup(ServiceGroupImpl serviceGroup);
}
