/*
 * PROJECT: mataereal at http://matereal.sourceforge.jp/
 * ----------------------------------------------------------------------------
 *
 * This file is part of matereal.
 * matereal is a Java toolkit that allows to operate home-robots easily.
 *
 * ----------------------------------------------------------------------------
 *
 * License version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * Variable length array.
 *
 * @author Jun KATO
 */
public class VariableLengthArray<T> implements Iterable<T> {
	final public static int INITIAL_SIZE = 10;

	/** Array for holding elements. */
	private Object[] elements = new Object[INITIAL_SIZE];

	/** Length of valid results. */
	private int validSize;

	/** Counter used in an iterator. */
	private int counterBias = 0;

	public VariableLengthArray() {
		// Do nothing.
	}

	public VariableLengthArray(VariableLengthArray<T> old) {
		if (old == null) {
			return;
		}

		// Copy the elements.
		realloc(old.size());
		for (T e : old) {
			elements[validSize ++] = e;
		}
	}

	/** Extend the holder size to the specified length. */
	public synchronized void realloc(int length) {
		if (length <= elements.length) {
			return;
		}
		elements = Arrays.copyOf(elements, length);
	}

	/** Returns the number of elements. */
	public int size() { return validSize; }

	/** Clear all elements. */
	public synchronized void clear() {
		Arrays.fill(elements, null);
		validSize = 0;
	}

	/** Put an element to the specified index. */
	public synchronized void put(T e, int index) {
		if (index < 0 || index + 1 >= elements.length) {
			throw new ArrayIndexOutOfBoundsException();
		}
		elements[index] = e;
	}

	/** Push an element. */
	public synchronized void push(T e) {
		if (validSize + 1 >= elements.length) {
			realloc(elements.length+1);
		}
		elements[validSize ++] = e;
	}

	/** Remove an element. */
	public synchronized void remove(int index) {
		for (int i = index; i < validSize-1;) {
			elements[i] = elements[++ i];
		}
		elements[-- validSize] = null;
		counterBias --;
	}

	/** Remove a specified element. */
	public synchronized boolean remove(T e) {
		if (e == null) {
			return false;
		}
		boolean found = false;
		for (int i = 0; i < validSize; i ++) {
			if (e.equals(elements[i])) {
				found = true;
			}
			if (found && i+1 < validSize) {
				elements[i] = elements[i+1];
			}
		}
		if (found) {
			elements[-- validSize] = null;
			counterBias --;
		}
		return found;
	}

	/** Get an element. */
	@SuppressWarnings("unchecked")
	public synchronized T get(int index) {
		if (index > validSize) {
			throw new ArrayIndexOutOfBoundsException();
		}
		return (T) elements[index];
	}

	/** Return this array as a List object. */
	public synchronized List<T> asList() {
		final ArrayList<T> list = new ArrayList<T>();
		for (T e : this) {
			list.add(e);
		}
		return list;
	}

	public Iterator<T> iterator() {
		return new VLAIterator();
	}

	private class VLAIterator implements Iterator<T> {
		private int counter;

		public VLAIterator() {
			synchronized (VariableLengthArray.this) {
				counter = -counterBias;
			}
		}
		public boolean hasNext() {
			synchronized (VariableLengthArray.this) {
				return counter + counterBias < validSize;
			}
		}

		@SuppressWarnings("unchecked")
		public T next() {
			synchronized (VariableLengthArray.this) {
				if (counter + counterBias >= validSize) {
					throw new NoSuchElementException();
				}
				return (T) elements[(counter ++) + counterBias];
			}
		}

		public void remove() {
			synchronized (VariableLengthArray.this) {
				VariableLengthArray.this.remove((counter --) + counterBias);
			}
		}
	}
}
